
/*************************************************************************
 *
 * thinkfan version 0.4 -- copyleft 01-2009, Victor Mataré
 *
 * A minimalist, lightweight fan control program for modern Thinkpads.
 * So far this is only confirmed to work on my T61p.
 *
 * This program is provided to you AS-IS. You can do with it whatever
 * you want. No warranties whatsoever. If this program steals your
 * car, kills your horse, smokes your dope or pees on your carpet...
 * - too bad, you're on your own.
 *
 *   !!!.. WARNING ..!!!.. WARNING ..!!!.. WARNING ..!!!................
 *   !!!
 *   !!!    This program could damage your hardware in any thinkable
 *   !!!    way. Most likely it will OVERHEAT YOUR CPU.
 *   !!!    There is NO SANITY CHECKING on the configuration.
 *   !!!    That means you can configure it to put your CPU on fire.
 *   !!!
 *   !!!    Don't run it if you don't understand what it does.
 *
 *
 *   You have been warned.
 *
 *************************************************************************/

#define VERSION "0.4"

#define _GNU_SOURCE

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <signal.h>
#include <errno.h>
#include "messages.c"

#define IBM_TEMP "/proc/acpi/ibm/thermal"
#define IBM_FAN "/proc/acpi/ibm/fan"
#define PID_FILE "/var/run/thinkfan.pid"

#define MAXERR 3 // maximum number of errors when trying to set fan speed

#define ERR_T_GET     	-1
#define ERR_FAN_INIT   	-2
#define ERR_CONF_NOFILE	-3
#define ERR_CONF_LOST	-4
#define ERR_CONF_RELOAD	-5
#define ERR_PIDFILE	   	-6
#define ERR_FANFILE    	-7

typedef struct {
	int level; // this is written into IBM_FAN to set fan level.
	int low;   // lower temperature for stepping down to previous level.
	int high;  // upper temperature determines when to step to the next level.
} thm_tuple;

typedef struct {
	thm_tuple *items;
	int max_idx;
} tf_config;

int quiet, nodaemon;
volatile int intr;
unsigned int sleeptime;
unsigned int watchdog_timeout;
unsigned int chk_sanity;
char *config_file;
float bias_level;

int run();

/*******************************************************************
 * get_temp() reads temperatures from /proc/acpi/ibm/thermal and
 * returns the highest one found.
 *******************************************************************/
int get_temp() {
	int i, res, retval=0, lasterr;
	int *t = calloc(16, sizeof(int));
	FILE *ibm_temp;

	memset(t, 0, 16);
	if ((ibm_temp = fopen(IBM_TEMP, "r")) == NULL) {
		lasterr = errno;
		if(nodaemon) perror(IBM_TEMP);
		syslog(LOG_ERR, IBM_TEMP ": %s", strerror(lasterr));
		free(t);
		return -1000;
	}
	res = fscanf(ibm_temp,
	"temperatures: %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i",
	 &t[0], &t[1], &t[2], &t[3], &t[4], &t[5], &t[6], &t[7], &t[8], &t[9],
	 &t[10], &t[11], &t[12], &t[13], &t[14], &t[15]);
	fclose(ibm_temp);
	if (res < 2) {
		syslog(LOG_ERR, MSG_ERR_T_NONE);
		if(nodaemon) fprintf(stderr, MSG_ERR_T_NONE);
		free(t);
		return -1000;
	}
	for (i=0; i < 16; i++)
		if (t[i] > retval) retval = t[i];
	free(t);
	return retval;
}


/*********************************************************
 * readconfig(char *fname) reads the config file and
 * returns a struct tf_config.
 * Non-matching lines are skipped.
 *********************************************************/
tf_config *readconfig(char* fname) {
	FILE *cfg_file;
	int i=0, num=0, lasterr, line_count=0;
	ssize_t line_len;
	size_t ll;
	tf_config *config;
	char *line=NULL;

	config = (tf_config *) malloc(sizeof(tf_config));
	config->items = (thm_tuple *) malloc(sizeof(thm_tuple));
	if ((cfg_file = fopen(fname, "r")) == NULL) {
		lasterr = errno;
		perror(fname);
		syslog(LOG_ERR, "%s: %s", fname, strerror(lasterr));
		return NULL;
	}
	while ((line_len = getline (&line, &ll, cfg_file)) >= 0) {
		line_count++;
		num = sscanf(line, " ( %d , %d , %d ) ",
		 &config->items[i].level, &config->items[i].low,
		 &config->items[i].high);
		if (num == 3) {
			i++;
			if ((config->items = realloc(config->items,
			 (i+1) * sizeof(thm_tuple))) == NULL) {
				perror("allocating memory for config");
				syslog(LOG_ERR, MSG_ERR_MEM);
				if(nodaemon) fprintf(stderr, MSG_ERR_MEM);
				return NULL;
			}
			// check for correct ordering of fan levels in config
			if ((i > 2) && (config->items[i-1].level <=
			 config->items[i-2].level)) {
				if (nodaemon && chk_sanity && !quiet)
					fprintf(stderr, MSG_ERR_CONF_ORDER);
				if (!quiet)
					syslog(LOG_ERR, MSG_ERR_CONF_ORDER);
				if (chk_sanity)
					return NULL;
			}
			if (config->items[i-1].high <= config->items[i-1].low) {
				if (nodaemon && chk_sanity && !quiet)
					fprintf(stderr, MSG_ERR_CONF_LOWHIGH);
				if (!quiet)
					syslog(LOG_ERR, MSG_ERR_CONF_LOWHIGH);
				if (chk_sanity)
					return NULL;
			}
			config->max_idx = i-1;
		}
		free(line);
		line = NULL;
	}
	if (line != NULL) {
		free(line);
		line = NULL;
	}

	fclose(cfg_file);
	if (i <= 0) {
		if(nodaemon) fprintf(stderr, MSG_ERR_CONF_PARSE);
		syslog(LOG_ERR, MSG_ERR_CONF_PARSE);
		free(config->items);
		free(config);
		return NULL;
	}
	return config;
}

/*********************************************************
 * Checks for fan_control support in thinkpad_acpi and
 * activates the fan watchdog.
 *********************************************************/
int init_fan() {
	char *line = NULL;
	size_t count = 0;
	FILE *ibm_fan;
	int lasterr, module_valid=0;

	if ((ibm_fan = fopen(IBM_FAN, "r+")) == NULL) {
		lasterr = errno;
		if (nodaemon) perror(IBM_FAN);
		syslog(LOG_ERR, IBM_FAN ": %s", strerror(lasterr));
		return -1;
	}
	while (getline(&line, &count, ibm_fan) != -1)
		if (!strncmp("commands:", line, 9)) module_valid = 1;
	if (!module_valid) {
		syslog(LOG_ERR, MSG_ERR_MODOPTS);
		if (nodaemon) fprintf(stderr, MSG_ERR_MODOPTS);
		return -2;
	}
	fprintf(ibm_fan, "watchdog %d\n", watchdog_timeout);
	fclose(ibm_fan);
	free(line);
	return 0;
}

/***********************************************************
 * Set fan speed.
 ***********************************************************/
int setfan(int level) {
	FILE *ibm_fan;
	int lasterr;

	if ((ibm_fan = fopen(IBM_FAN, "r+")) == NULL) {
		lasterr = errno;
		if (nodaemon) perror(IBM_FAN);
		syslog(LOG_ERR, IBM_FAN ": %s", strerror(lasterr));
		return -1;
	}
	int ret = fprintf(ibm_fan, "level %d\n", level);
	fclose(ibm_fan);
	if (ret != 8) {
		if (nodaemon) fprintf(stderr, MSG_ERR_FAN_WRITE);
		syslog(LOG_ERR, MSG_ERR_FAN_WRITE);
		return -2;
	}
	return 0;
}

/***********************************************************
 * This is the main routine which periodically checks
 * temperatures and adjusts the fan according to config.
 ***********************************************************/
int fancontrol(tf_config *config) {
	int last_temp, temp, lvl_idx=0, errcnt=0, bias=0, diff=0, b_temp;
	int wt = watchdog_timeout, st = sleeptime;

	if (config == NULL) {
		if(nodaemon) fprintf(stderr, MSG_ERR_CONF_LOST);
		syslog(LOG_ERR, MSG_ERR_CONF_LOST);
		return -1;
	}
	else if (!quiet) {
		int i;
		if (nodaemon)
			fprintf(stderr, MSG_INF_CONFIG);
		syslog(LOG_NOTICE, MSG_INF_CONFIG);
		for (i=0; i <= config->max_idx; i++) {
			syslog(LOG_INFO, MSG_INF_CONF_ITEM,
			 config->items[i].level, config->items[i].low,
			 config->items[i].high);
			if (nodaemon)
			 fprintf(stderr, MSG_INF_CONF_ITEM,
			 config->items[i].level, config->items[i].low,
			 config->items[i].high);
		}
	}
	if (init_fan()) {
		if(nodaemon) fprintf(stderr, MSG_ERR_FAN_INIT);
		syslog(LOG_ERR, MSG_ERR_FAN_INIT);
		return -2;
	}

// Set initial fan level:
	lvl_idx = config->max_idx;
	temp = get_temp();
	if (temp < 1) {
		if(nodaemon) fprintf(stderr, MSG_ERR_T_GET);
		syslog(LOG_ERR, MSG_ERR_T_GET);
		return -3;
	}
	while ((temp <= config->items[lvl_idx].low) \
	 && (lvl_idx > 0)) lvl_idx--;
	if (setfan(config->items[lvl_idx].level)) errcnt++;

/**********************************************
 * Main loop. This is the actual fan control.
 **********************************************/
	while(!intr) { // intr is set by the signal handler
		last_temp = temp; // detect temperature jumps
		sleep(st); // st is the variable sleeptime
		temp = get_temp();

		// If temperature increased by more than 2 °C since the
		// last cycle, we try to react quickly.
		diff = temp - last_temp;
		if (diff >= 2) {
			bias = (st * (diff-1)) * bias_level;
			if (st > 2) st = 2;
		}
		else if (st < sleeptime) st++;
		b_temp = temp + bias;
		if (b_temp >= config->items[lvl_idx].high) {
			while ((b_temp >= config->items[lvl_idx].high) \
			 && (lvl_idx <= config->max_idx)) lvl_idx++;
			if (setfan(config->items[lvl_idx].level))
				if (++errcnt >= MAXERR) return errcnt;
#ifdef DEBUG
			syslog(LOG_DEBUG, MSG_DBG_T_STAT);
			if (nodaemon) fprintf(stderr, MSG_DBG_T_STAT);
#endif
		}
		if (b_temp <= config->items[lvl_idx].low) {
			while ((b_temp <= config->items[lvl_idx].low) \
			 && (lvl_idx > 0)) lvl_idx--;
			if (setfan(config->items[lvl_idx].level))
				if (++errcnt >= MAXERR) return errcnt;
			st = sleeptime;
#ifdef DEBUG
			syslog(LOG_DEBUG, MSG_DBG_T_STAT);
			if (nodaemon) fprintf(stderr, MSG_DBG_T_STAT);
#endif
		}

// slowly reduce the bias
		if (bias > 0) bias -= (bias_level + 0.1f) * 4;
		else (bias = 0);
// Write current fan level to IBM_FAN one cycle before watchdog timeout
// ends, to let it know we're alive.
		if ( (wt -= st) <= st ) {
			if (setfan(config->items[lvl_idx].level))
				if (++errcnt >= MAXERR) return errcnt;
			wt = watchdog_timeout;
		}

	}
	return 0;
}

void sigHandler(int signum) {
	switch(signum) {
	case SIGHUP:
		intr = signum;
		break;
	case SIGINT:
	case SIGTERM:
		intr = signum;
		break;
	}
}


/***************************************************************
 * Main function:
 * Scan for arguments, set options and initialize signal handler
 ***************************************************************/
int main(int argc, char **argv) {
	int opt, ret, fork_ret;
	char *invalid = "";
	struct sigaction handler;

	sleeptime = 5;
	quiet = 0;
	intr = 0;
	chk_sanity = 1;
	bias_level = 0.5f;
	ret = 0;
	config_file = "/etc/thinkfan.conf";
	nodaemon=0;

	openlog("thinkfan", LOG_CONS, LOG_USER);
	syslog(LOG_INFO, "thinkfan " VERSION " starting...");

	while ((opt = getopt(argc, argv, "c:s:b:hnqD")) != -1) {
		switch(opt) {
		case 'h':
			fprintf(stderr, MSG_USAGE);
			return 0;
			break;
		case 'n':
			nodaemon = 1;
			break;
		case 'c':
			config_file = optarg;
			break;
		case 's':
			sleeptime = (unsigned int) strtol(optarg, &invalid, 0);
			if (*invalid != 0) {
				fprintf(stderr, MSG_ERR_OPT_S MSG_USAGE);
				return 1;
			}
			break;
		case 'q':
			quiet = 1;
			break;
		case 'D':
			chk_sanity = 0;
			break;
		case 'b':
			bias_level = strtof(optarg, &invalid);
			if (*invalid != 0) {
				fprintf(stderr, MSG_ERR_OPT_B MSG_USAGE);
				return 1;
			}
			if (bias_level >= 0 && bias_level <= 20)
				bias_level = 0.1f * bias_level;
			else {
				fprintf(stderr, MSG_ERR_OPT_B MSG_USAGE);
				return 1;
			}
			break;
		default:
			fprintf(stderr, MSG_USAGE);
			return 1;
		}
	}

	if (sleeptime > 15) {
		if (chk_sanity) {
			fprintf(stderr, MSG_WRN_SLEEPTIME_15);
			fprintf(stderr, MSG_INF_SANITY);
			return 1;
		}
		else if (!quiet) {
			fprintf(stderr, MSG_WRN_SLEEPTIME_15);
			fprintf(stderr, MSG_INF_INSANITY);
		}
	}
	else if (sleeptime < 1) {
		if (chk_sanity) {
			fprintf(stderr, MSG_WRN_SLEEPTIME_1);
			fprintf(stderr, MSG_INF_SANITY);
			return 1;
		}
		else if (!quiet) {
			fprintf(stderr, MSG_WRN_SLEEPTIME_1);
			fprintf(stderr, MSG_INF_INSANITY);
		}
	}

	memset(&handler, 0, sizeof(handler));
	handler.sa_handler = sigHandler;
	if (sigaction(SIGHUP, &handler, NULL) \
	 || sigaction(SIGINT, &handler, NULL) \
	 || sigaction(SIGTERM, &handler, NULL)) perror("sigaction");
	watchdog_timeout = sleeptime * 6;

	if (nodaemon) ret = run();
	else {
		if ((fork_ret = fork()) == 0) ret = run();
		else if (fork_ret < 0) {
			perror("fork()");
			ret = 1;
		}
		else if (!quiet)
			fprintf(stderr, "Daemon PID: %d", fork_ret);
	}
	return ret;
}


/********************************************************************
 * Outer loop. Handles signal conditions, runtime cleanup and config
 * reloading
 ********************************************************************/
int run() {
	int lasterr, ret = 0;
	tf_config *config, *newconfig=NULL;
	FILE *pidfile;

	if ((config = readconfig(config_file)) == NULL) {
		if (nodaemon) fprintf(stderr, MSG_ERR_CONF_NOFILE);
		syslog(LOG_ERR, MSG_ERR_CONF_NOFILE);
		return -1;
	}

	if ((pidfile = fopen(PID_FILE, "w+")) == NULL) {
		lasterr = errno;
		if (nodaemon) perror(PID_FILE);
		syslog(LOG_ERR, PID_FILE ": %s", strerror(lasterr));
		return -2;
	}
	fprintf(pidfile, "%d", getpid());
	fclose(pidfile);

	while ((intr < SIGINT) && !ret) {
		intr = 0;
		ret = fancontrol(config);
		if (ret) {
			if (ret > 0) { //ret > 0 represents a fan error count
				syslog(LOG_ERR, MSG_ERR_ERRCOUNT);
				if (nodaemon) fprintf(stderr, MSG_ERR_ERRCOUNT);
			}
			break;
		}
		else if (intr < SIGINT) {
			syslog(LOG_DEBUG, MSG_DBG_CONF_RELOAD);
			if (nodaemon)
				fprintf(stderr, MSG_DBG_CONF_RELOAD);
			if ((newconfig = readconfig(config_file)) != NULL) {
				free(config->items);
				free(config);
				config = newconfig;
			}
			else {
				fprintf(stderr, MSG_ERR_CONF_RELOAD);
				syslog(LOG_ERR, MSG_ERR_CONF_RELOAD);
			}
		}
		else {
			if (nodaemon)
				fprintf(stderr, MSG_INF_TERM);
			syslog(LOG_WARNING, MSG_INF_TERM);
			break;
		}
	}
	free(config->items);
	free(config);
	FILE *fan;
	if ((fan = fopen(IBM_FAN, "r+")) == NULL) {
		int lasterr = errno;
		perror(IBM_FAN);
		syslog(LOG_ERR, "Opening " IBM_FAN ": %s", strerror(lasterr));
		return ERR_FANFILE;
	}
	fprintf(fan, "level auto\n");
	fclose(fan);
	return ret;
}

